import { MessageTransmissionTransmitter } from "./msg.transmission.transmitter";
import { MessageTransmissionReceiver } from "./msg.transmission.receiver";
import { AdapterManager } from "../adapters/adapter.manager";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionRequestResponse } from "./msg.transmission.request-response";
import { filter, Observable, Observer, Subject, Subscription, Unsubscribable } from "rxjs";
import ConsoleLogger from "../utils/log.utils";
import { MessageTransmissionManagerBase } from "../base/msg.transmission.manager.base";
import { AdapterInterface, ClientObject, GeneralEvent, TransmissionInterface } from "../interface/interface";

export class MessageTransmissionManager extends MessageTransmissionManagerBase {
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionManager`, ['managers'])

    constructor(event: Subject<GeneralEvent<any>>, browserEnv?: boolean) {
        super()
        if (browserEnv) this.browserEnv = browserEnv
        this.console.log({ message: `Contructing self... ${this.browserEnv ? `is receiving end` : `is not browser env`}` })
        this.event = event
        // Subscribe for adapterManager and it's relevent event
        this.adapterManager = new AdapterManager(event, this.browserEnv)
        // this.adapterManager.subscribeForAdapters().subscribe(this.event)
        this.adapterManager.subscribeForAdapters().subscribe((adapterEvent: GeneralEvent<AdapterInterface>) => {
            this.event.next(adapterEvent)
        })
    }

    public subscribeForTransmission(): Observable<GeneralEvent<TransmissionInterface>> {
        return new Observable((observer: Observer<GeneralEvent<TransmissionInterface>>) => {
            const subscription: Subscription = this.event.pipe(
                filter(event => event.type === `Transport Event`),
                filter(event => event.event === `New Client` || event.event === `New Server`)
            ).subscribe((event: GeneralEvent<ClientObject>) => {
                // get all adapters for all the connection
                let transmission: TransmissionInterface | undefined = this.instantiateTransmissionComponents(event.data, this.event)
                this.console.log({ message: `Passing this transmission for client:${transmission.clientId} to global event bus.` })
                if (transmission) {
                    this.event.next({
                        id: uuidv4(),
                        type: `Transmission Event`,
                        event: 'New Transmission',
                        date: new Date(),
                        data: transmission
                    })
                }
            })

            // Clean up on unsubscription
            return () => {
                subscription.unsubscribe();
            };
        })
    }

    protected instantiateTransmissionComponents(clientObj: ClientObject, eventRef: Subject<GeneralEvent<any>>): TransmissionInterface {
        let receiverInstance: MessageTransmissionReceiver = new MessageTransmissionReceiver(clientObj.clientId, eventRef)
        let transmitterInstance: MessageTransmissionTransmitter = new MessageTransmissionTransmitter(clientObj.clientId, eventRef)
        let requestResponseInstance: MessageTransmissionRequestResponse = new MessageTransmissionRequestResponse(transmitterInstance, receiverInstance, eventRef)
        let transmissionObj: TransmissionInterface = {
            clientId: clientObj.clientId,
            transmitter: transmitterInstance,
            receiver: receiverInstance,
            requestResponse: requestResponseInstance,
            clientInfo: clientObj
        }

        return transmissionObj
    }

}


