import { MessageTransmissionBase } from "../base/msg.transmission.base";
import { v4 as uuidv4 } from 'uuid'
import { BehaviorSubject, distinctUntilChanged, filter, map, Observable, Subject, Subscription } from "rxjs";
import { RetransmissionService } from "../utils/retransmission.service";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { AdapterInterface, ConnectionState, FisMessage, GeneralEvent, MessageTransmitterInterface, TransmitterAdapterInterface, TransportMessage } from "../interface/interface";
import { error } from "console";

/* Take in all the messages that needs to be transported, and divide them accordingly. So the connector instances will do just that
connectors or adapters will have their own identifier*/
export class MessageTransmissionTransmitter extends MessageTransmissionBase implements MessageTransmitterInterface {
    private connectionStateEvent: BehaviorSubject<ConnectionState> = new BehaviorSubject<ConnectionState>('OFFLINE')
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionTransmitter`, ['transmission'])
    private messageToBeBuffered!: Subject<FisMessage | WrappedMessage>
    private messageToBeTransmitted!: Subject<WrappedMessage>
    private buffer!: RetransmissionService;
    private currentAdapter!: TransmitterAdapterInterface

    constructor(clientId: string, event: Subject<GeneralEvent<any>>) {
        super()
        this.console.log({ message: `Constructing Transmitter Transmission with ${clientId}` })
        this.clientId = clientId
        this.event = event
        this.messageToBeTransmitted = new Subject()
        this.messageToBeBuffered = new Subject()
        this.buffer = new RetransmissionService()
        this.handleAdapters(this.event.asObservable())
        this.setupBuffer()

        // special case just for http in case of server/client disconnected, the unsent msg will be flushed back into messageToBeBuffered
       
    }

    public emit(message: FisMessage): void {
        this.console.log({ message: `${this.connectionStateEvent.getValue() == 'ONLINE' ? `Transmitting message` : `Buffering message`}` })
        this.messageToBeBuffered.next(message)
    }

    private setupBuffer(): void {
        this.console.log({ message: `Setting up Retransmission Service...` })
        this.event.pipe(
            filter(event => event.type === `Transport Event`),
            filter(event => event.event == 'Client Disconnected' || event.event == 'Client Re-connected' || event.event == 'Client Connected' || event.event == 'Server Disconnected' || event.event == 'Server Connected'),
            filter(event => event.data.clientId == this.clientId),
            map(event => {
                if (event.event == 'Client Disconnected' || event.event == 'Server Disconnected') {
                    return 'OFFLINE'
                } else {
                    return `ONLINE`
                }
            }),
            distinctUntilChanged()
        ).subscribe((signal: ConnectionState) => {
            this.connectionStateEvent.next(signal)
            if (signal == 'OFFLINE') this.console.error({ message: `${this.clientId} disconnected` })
            if (signal == 'ONLINE') this.console.log({ message: `${this.clientId} connected and ready to go` })
        })
        this.buffer.implementRetransmission(this.messageToBeBuffered, this.connectionStateEvent.asObservable(), true)
        // automatically subscribe to allow released bffered messages to be released
        this.buffer.returnSubjectForBufferedItems().subscribe((bufferedMessage: WrappedMessage) => {
            // need to work with wrapped messages
            this.console.log({ message: `Releasing ${bufferedMessage.thisMessageID}` });
            if (this.currentAdapter) {
                // this.currentAdapter.emit(bufferedMessage)
                this.messageToBeTransmitted.next(bufferedMessage)
            } else {
                this.messageToBeBuffered.next(bufferedMessage)
                this.console.error({ message: `Adapter is not set. Please ensure adapters are ready.` })
            }
        })
    }

    private handleAdapters(adaptersEvent: Observable<GeneralEvent<any>>): void {
        // hardcode it to use the first adapter added for now. Haven't decide on the logic to dynamically switch them adapters
        adaptersEvent.pipe(
            filter(event => event.type === `Adapter Event`),
            filter(event => event.event === `New Adapter`),
            map(event => { return event.data }),
            filter((adapter: AdapterInterface) => adapter.role === `Transmitter`),
            map(adapter => { return adapter as TransmitterAdapterInterface })
        ).subscribe({
            next: (adapter: TransmitterAdapterInterface) => {
                this.adapters.push(adapter)
                if (!this.currentAdapter) {
                    this.currentAdapter = adapter as TransmitterAdapterInterface
                    this.console.log({ message: `Setting Current adapter = ${this.currentAdapter.adapterId}` })
                } else {
                    this.console.log({ message: `Already have existing transmitting adapter. Current adapter = ${this.currentAdapter.adapterId}` })
                }
            },
            error: error => this.console.error({ message: 'Observer Error', details: error })
        })
    }

    private uniqueHandlerToFlushUnsentMessages(event: Observable<GeneralEvent<any>>): void {
        event.pipe(
            filter(event => event.event == 'Re-Flush'),
            filter(event => event.data.clientId == this.clientId),
        ).subscribe((event: GeneralEvent<any>) => {
            this.console.log({ message: `${this.connectionStateEvent.getValue() == 'ONLINE' ? `Transmitting ${((event.data.payload as TransportMessage).payload as WrappedMessage).thisMessageID}` : `Buffering ${((event.data.payload as TransportMessage).payload as WrappedMessage).thisMessageID}`}` })
            this.messageToBeBuffered.next(((event.data.payload as TransportMessage).payload as WrappedMessage))
        })
    }

}